/**
 * @vue-async/module-loader@3.1.1
 */

'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var utils = require('@vue-async/utils');
var Vue = _interopDefault(require('vue'));

/**
 * Copyright (c) 2014-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

/**
 * Similar to invariant but only logs a warning if the condition is not met.
 * This can be used to log issues in development environments in critical
 * paths. Removing the logging code for production environments will keep the
 * same logic and follow the same code paths.
 */

var __DEV__ = process.env.NODE_ENV !== 'production';

var warning = function() {};

if (__DEV__) {
  var printWarning = function printWarning(format, args) {
    var len = arguments.length;
    args = new Array(len > 1 ? len - 1 : 0);
    for (var key = 1; key < len; key++) {
      args[key - 1] = arguments[key];
    }
    var argIndex = 0;
    var message = 'Warning: ' +
      format.replace(/%s/g, function() {
        return args[argIndex++];
      });
    if (typeof console !== 'undefined') {
      console.error(message);
    }
    try {
      // --- Welcome to debugging React ---
      // This error was thrown as a convenience so that you can use this stack
      // to find the callsite that caused this warning to fire.
      throw new Error(message);
    } catch (x) {}
  };

  warning = function(condition, format, args) {
    var len = arguments.length;
    args = new Array(len > 2 ? len - 2 : 0);
    for (var key = 2; key < len; key++) {
      args[key - 2] = arguments[key];
    }
    if (format === undefined) {
      throw new Error(
          '`warning(condition, format, ...args)` requires a warning ' +
          'message argument'
      );
    }
    if (!condition) {
      printWarning.apply(null, [format].concat(args));
    }
  };
}

var warning_1 = warning;

/**
 * eventBus
 */
// eslint-disable-next-line @typescript-eslint/no-unused-vars
function createEventBus (Vue) {
    let events = {};
    return {
        /**
         * 发射事件
         * @param {string} eventName 事件名称
         * @param {any} payload 事件载荷
         */
        emit(eventName, payload) {
            if (!events[eventName]) {
                return warning_1(process.env.NODE_ENV === 'production', `[moduleLoader] event "${eventName}" is triggered, but no listeners，nothing will be happening.`);
            }
            for (const func of events[eventName].values()) {
                func(payload);
            }
        },
        /**
         * 事件监听
         * @param eventName 事件名称
         * @param handler 处理事件的方法
         */
        on(eventName, handler) {
            if (typeof handler === 'function') {
                if (!events[eventName])
                    events[eventName] = new Set();
                events[eventName].add(handler);
            }
            else {
                warning_1(false, `[moduleLoader] handler must be a function, but recived ${typeof handler}`);
            }
        },
        /**
         * 取消事件监听
         * @param eventName 事件名称
         * @param handler 处理事件的方法
         */
        off(eventName, handler) {
            if (!handler) {
                return warning_1(false, '[moduleLoader] handler must be a valid event function');
            }
            else if (!events.hasOwnProperty(eventName)) {
                return warning_1(process.env.NODE_ENV === 'production', `[moduleLoader] no event name of "${eventName}" is on listening`);
            }
            events[eventName].delete(handler);
        },
        /**
         * 清理事件总线
         */
        clear() {
            events = {};
        },
        /**
         * 获取当前事件总线详情
         */
        getEvents() {
            return events;
        },
    };
}

/**
 * @author Hubert
 * fork from https://github.com/kuitos/import-html-entry/blob/master/src/utils.js
 */
const isIE11 = typeof navigator !== 'undefined' && navigator.userAgent.indexOf('Trident') !== -1;
function shouldSkipProperty(global, p) {
    if (!global.hasOwnProperty(p) || (!isNaN(p) && parseInt(p) < global.length))
        return true;
    if (isIE11) {
        // https://github.com/kuitos/import-html-entry/pull/32，最小化 try 范围
        try {
            return global[p] && global[p].parent === window;
        }
        catch (err) {
            return true;
        }
    }
    else {
        return false;
    }
}
// safari unpredictably lists some new globals first or second in object order
let firstGlobalProp, secondGlobalProp, lastGlobalProp;
function getGlobalProp(global) {
    let cnt = 0;
    let lastProp;
    let hasIframe = false;
    for (const p in global) {
        if (shouldSkipProperty(global, p))
            continue;
        // 遍历 iframe，检查 window 上的属性值是否是 iframe，是则跳过后面的 first 和 second 判断
        for (let i = 0; i < window.frames.length && !hasIframe; i++) {
            const frame = window.frames[i];
            if (frame === global[p]) {
                hasIframe = true;
                break;
            }
        }
        if (!hasIframe && ((cnt === 0 && p !== firstGlobalProp) || (cnt === 1 && p !== secondGlobalProp)))
            return p;
        cnt++;
        lastProp = p;
    }
    if (lastProp !== lastGlobalProp)
        return lastProp;
    return;
}
function noteGlobalProps(global) {
    // alternatively Object.keys(global).pop()
    // but this may be faster (pending benchmarks)
    firstGlobalProp = secondGlobalProp = undefined;
    for (const p in global) {
        if (shouldSkipProperty(global, p))
            continue;
        if (!firstGlobalProp)
            firstGlobalProp = p;
        else if (!secondGlobalProp)
            secondGlobalProp = p;
        lastGlobalProp = p;
    }
    return lastGlobalProp;
}
/** 加载 entry 脚本 */
function execScript(entry, proxy = window) {
    return new Promise((resolve, reject) => {
        noteGlobalProps(proxy);
        const script = proxy.document.createElement('script');
        script.src = entry;
        script.onload = () => {
            const propName = getGlobalProp(proxy);
            const exports = propName ? proxy[propName] || {} : {};
            resolve(exports);
        };
        script.onerror = (err) => {
            warning_1(process.env.NODE_ENV === 'production', `[moduleLoader] script had a problem to create, entry：${entry}`);
            reject(new Error(`script load error, error: ${err}`));
        };
        proxy.document.body.appendChild(script);
    });
}
/** 加载 styles */
function execStyles(styles, styleFor) {
    if (styles.length) {
        return Promise.all(styles.map((href) => {
            return new Promise((resolve) => {
                const link = document.createElement('link');
                link.rel = 'stylesheet';
                link.type = 'text/css';
                link.href = href;
                styleFor && (link.dataset.styleFor = styleFor);
                link.onload = resolve;
                link.onerror = resolve;
                document.getElementsByTagName('head')[0].appendChild(link);
            });
        }));
    }
    return Promise.resolve();
}

/**
 * moduleLoader
 */
const noop = () => { };
function promisify(promise) {
    if (promise && promise instanceof Promise && typeof promise.then === 'function') {
        return promise;
    }
    return Promise.resolve(promise);
}
/** 格式化模块配置 */
function formatModules(config, errorCallback = () => { }) {
    if (Array.isArray(config)) {
        return config.reduce((prev, curr) => [...prev, ...formatModules(curr)], []);
    }
    else if (utils.isPlainObject(config)) {
        const _modules = config;
        // no lib name
        if (_modules.entry && typeof _modules.entry === 'string') {
            return [
                {
                    ...config,
                    styles: _modules.styles ? (!Array.isArray(_modules.styles) ? [_modules.styles] : _modules.styles) : [],
                },
            ];
        }
        else {
            return Object.keys(_modules).map((moduleName) => typeof _modules[moduleName] === 'string'
                ? {
                    entry: _modules[moduleName],
                    moduleName,
                }
                : {
                    ..._modules[moduleName],
                    styles: _modules[moduleName].styles
                        ? !Array.isArray(_modules[moduleName].styles)
                            ? [_modules[moduleName].styles]
                            : _modules[moduleName].styles
                        : [],
                    moduleName,
                });
        }
    }
    else if (utils.isFunction(config)) {
        return [config];
    }
    else {
        errorCallback(`not support "${typeof config}" type module`, config);
        return [];
    }
}
/** 校验子应用导出的 生命周期 对象是否正确 */
function validateExportLifecycle(exports) {
    // export default function(){} 兼容处理
    const _export = (exports && exports.default) || exports;
    if (utils.isFunction(_export)) {
        return true;
    }
    else {
        // 生命周期
        // todo: 延迟加载
        // const { bootstrap ，mount, unmount} = exports ?? {};
        // return isFunction(bootstrap) && isFunction(mount) && isFunction(unmount);
        const { bootstrap } = exports || {};
        return utils.isFunction(bootstrap);
    }
}
/** 获取生命周期对象 */
function getLifecyclesFromExports(scriptExports, moduleName, global) {
    if (validateExportLifecycle(scriptExports)) {
        const _export = (scriptExports && scriptExports.default) || scriptExports;
        return utils.isFunction(_export) ? { bootstrap: _export } : scriptExports;
    }
    warning_1(process.env.NODE_ENV === 'production', `[moduleLoader] lifecycle not found from ${moduleName} entry exports, fallback to get from window['${moduleName}']`);
    // fallback to global variable who named with ${moduleName} while module exports not found
    const globalVariableExports = global.exports ? global.exports[moduleName] : global[moduleName];
    if (validateExportLifecycle(globalVariableExports)) {
        const _export = (globalVariableExports && globalVariableExports.default) || globalVariableExports;
        return utils.isFunction(_export) ? { bootstrap: _export } : globalVariableExports;
    }
    throw new Error(`[moduleLoader] You need to export lifecycle functions in ${moduleName} entry`);
}
var createModuleLoader = (Vue, status) => {
    return function loader(config, options = {}) {
        const { sync, onLoading = noop, onLoaded = noop, onError = (name, error) => {
            warning_1(false, `An error occurred while loading module "${name}", Error: ${error.message}`);
        }, success, error, } = options;
        status.current = false;
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        const _self = this;
        const _modules = formatModules(config, error);
        // 按顺序同步执行
        if (sync === true) {
            return new Promise((resolve) => schedule(0, () => {
                status.current = true;
                success ? success() : resolve();
            }));
        }
        else {
            return new Promise((resolve) => Promise.all(_modules.map((module) => exec(module))).then(() => {
                status.current = true;
                success ? success() : resolve();
            }));
        }
        async function exec(module) {
            // local module
            if (utils.isFunction(module)) {
                const result = module.call(_self, Vue);
                if (result && result instanceof Promise) {
                    return result;
                }
                return Promise.resolve();
            }
            // remote module
            else {
                const { moduleName, entry, styles = [], args } = module;
                // server render
                // if (Vue.prototype.$isServer) {
                //   const global = ssr.createSandbox();
                //   return ssr
                //     .execScript(entry, global)
                //     .then((scriptExports: any) => {
                //       const { bootstrap } = getLifecyclesFromExports(scriptExports, moduleName, global);
                //       bootstrap.call(_self, Vue, args);
                //     })
                //     .catch((err: Error) => {
                //       // 异常不阻止当前执行，error 中处理
                //       try {
                //         error(err.message, module);
                //       } catch {}
                //     });
                // } else {
                // todo: load in sandbox
                const global = window;
                if (!global.Vue) {
                    global.Vue = Vue;
                }
                try {
                    // load styles
                    await execStyles(styles, moduleName);
                    // exec script
                    const scriptExports = await execScript(entry, global);
                    await promisify(onLoading(moduleName));
                    const { bootstrap } = getLifecyclesFromExports(scriptExports, moduleName, global);
                    await promisify(bootstrap.call(_self, Vue, args));
                    await promisify(onLoaded(moduleName));
                }
                catch (err) {
                    // 异常不阻止当前执行，error 中处理
                    try {
                        await promisify(onError(moduleName, err));
                        // 方法已过时
                        if (error && typeof error === 'function') {
                            // eslint-disable-next-line no-console
                            console.warn('"error" function is deprecated, using "onError" function!');
                            error(err.message, module);
                        }
                    }
                    catch { }
                }
                // }
            }
        }
        function schedule(index, resolve) {
            if (index < _modules.length) {
                exec(_modules[index]).then(() => {
                    schedule(index + 1, resolve);
                });
            }
            else {
                resolve();
            }
        }
    };
};

/**
 * componentLoader
 */
/** 验证组件导出是否正确 */
function validateExportComponent(exports) {
    const _export = (exports && exports.default) || exports;
    if (utils.isPlainObject(_export)) {
        return (Object.keys(_export).length > 0 &&
            (utils.hasOwn(_export, 'template') ||
                utils.hasOwn(_export, 'render') || // component definition
                utils.isFunction(_export)) // Vue.extend
        // hasOwn(_export, 'component') || // AsyncComponentFactory
        //   isFunction(_export)) // AsyncComponentPromise
        );
    }
    else {
        return _export instanceof Vue;
    }
}
/** 获取组件 */
function getComponentFromExport(scriptExports, componentName, global) {
    if (validateExportComponent(scriptExports)) {
        return (scriptExports && scriptExports.default) || scriptExports;
    }
    warning_1(false, `[moduleLoader] component not found from ${componentName} entry exports, fallback to get from window['${componentName}']`);
    // fallback to global variable who named with ${moduleName} while module exports not found
    const globalVariableExports = global[componentName];
    if (validateExportComponent(globalVariableExports)) {
        return (globalVariableExports && globalVariableExports.default) || globalVariableExports;
    }
    throw new Error(`[moduleLoader] You need to export vue component in ${componentName} entry`);
}
// eslint-disable-next-line @typescript-eslint/no-unused-vars
var createComponentLoader = (Vue) => {
    return function loader(componentName, src, styles) {
        // if (Vue.prototype.$isServer) {
        //   const global = ssr.createSandbox();
        //   return ssr.execScript(src, global).then((scriptExports) => {
        //     return getComponentFromExport(scriptExports, componentName, global);
        //   });
        // } else {
        const global = window;
        if (!global.Vue) {
            global.Vue = Vue;
        }
        // load styles
        if (styles) {
            if (typeof styles === 'string') {
                styles = [styles];
            }
            execStyles(styles, componentName);
        }
        // exec script
        return execScript(src, global).then((scriptExports) => {
            return getComponentFromExport(scriptExports, componentName, global);
        });
        // }
    };
};

function install(Vue) {
    if (install.installed)
        return;
    install.installed = true;
    // 监听模块加载的情况，处理路由404问题
    const vm = new Vue({
        data() {
            return {
                status: {
                    current: false,
                },
            };
        },
    });
    const properties = {
        $moduleLoader: {
            value: createModuleLoader(Vue, vm.status),
            writable: false,
            enumerable: true,
            configurable: true,
        },
        $componentLoader: {
            value: createComponentLoader(Vue),
            writable: false,
            enumerable: true,
            configurable: true,
        },
        $eventBus: {
            value: createEventBus(),
            writable: false,
            enumerable: true,
            configurable: true,
        },
    };
    /**
     * 注入到 Vue 实例
     */
    Object.defineProperties(Vue.prototype, properties);
    /**
     * router注入（当路由被注册注入时，必须调用$moduleLoaser至少一次，否则 status 状态无法变更造成路由被阻止）
     * @param router
     */
    const routerInject = (router) => {
        return router.beforeEach((to, from, next) => {
            if (!to.name || to.name === '404' || to.name === 'page-not-found' || to.name === 'not-found' || to.path === '*') {
                // 模块已经被加载完成, 但由于在其之前添加的 beforeEach 阻止时间过长，vm.$watch还没开始监听
                if (vm.status.current) {
                    return resolveRoute(to, from, next);
                }
                // 当 status 为 false 时watch状态
                vm.$watch(() => vm.status.current, (newVal, oldVal) => {
                    // 从false => true状态
                    if (newVal && !oldVal) {
                        resolveRoute(to, from, next);
                    }
                });
            }
            else {
                next();
            }
        });
        // 解决动态路由404问题
        function resolveRoute(to, from, next) {
            const fullPath = to.redirectedFrom || to.fullPath;
            const { resolved, location } = router.resolve(fullPath);
            // 在加载完组件后resolve的地址与原来要跳转的地址不一致时跳转
            // 以免造成404死循环
            if (resolved.name !== to.name) {
                next(location);
            }
            else {
                next();
            }
        }
    };
    const _init = Vue.prototype._init;
    /**
     * 从 Vue root.options 中获取 router | store 实例
     */
    Vue.prototype._init = function (options = {}) {
        // router beforeEach 需要在 beforeCreate 之前添加，才能在页面强制刷新时第一次生效
        if (options.router) {
            routerInject(options.router);
        }
        _init.call(this, options);
    };
    // 扩展 new ModuleLoader().load() 在 Vue 实例化之前加载模块
    Object.defineProperties(this.prototype, {
        load: properties.$moduleLoader,
        loadComponent: properties.$componentLoader,
        eventBus: properties.$eventBus,
    });
    // Used to avoid multiple mixins being setup
    // when in dev mode and hot module reload
    // https://github.com/vuejs/vue/issues/5089#issuecomment-284260111
    if (Vue.$__module_loader_installed__)
        return;
    // eslint-disable-next-line @typescript-eslint/camelcase
    Vue.$__module_loader_installed__ = true;
    Vue.mixin({
        beforeCreate() {
            const options = this.$options;
            if (options.moduleLoader) {
                options.moduleLoader.init(this, options.ssrContext);
                this.$moduleLoadManager = Vue.observable(options.moduleLoader.framework);
            }
            else {
                this.$moduleLoadManager = (options.parent && options.parent.$moduleLoadManager) || this;
            }
        },
    });
}

class ModuleLoader {
    constructor(options) {
        this._framework = {};
        options &&
            Object.entries(options).map(([key, value]) => {
                if (!utils.hasOwn(this._framework, key)) {
                    this._framework[key] = value;
                }
                else {
                    warning_1(process.env.NODE_ENV === 'production', `${key} has been defined`);
                }
            });
    }
    get framework() {
        return this._framework;
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    init(root, ssrContext) {
        // 默认方法
        !utils.hasOwn(this._framework, 'addRoutes') && (this._framework.addRoutes = this._createAddRoutes(root));
    }
    _createAddRoutes(root) {
        return (routes) => {
            if (root.$router) {
                root.$router.addRoutes(routes);
            }
            else {
                warning_1(process.env.NODE_ENV === 'production', 'vm.$router is not exists!');
            }
        };
    }
}
ModuleLoader.install = install;
ModuleLoader.installed = false;
ModuleLoader.version = '3.1.1';

// function sortBy(name: string, order: 'ASC' | 'DESC' = 'ASC') {
//   return function(o: Dictionary<any>, p: Dictionary<any>) {
//     let a, b;
//     if (typeof o === 'object' && typeof p === 'object' && o && p) {
//       a = o[name];
//       b = p[name];
//       if (a === b) {
//         return 0;
//       }
//       if (typeof a === typeof b) {
//         const result = a < b ? -1 : 1;
//         return order === 'DESC' ? ~result + 1 : result;
//       }
//       const result = typeof a < typeof b ? -1 : 1;
//       return order === 'DESC' ? ~result + 1 : result;
//     } else {
//       throw 'error';
//     }
//   };
// }
class POHook {
    constructor() {
        this.callbacks = {};
        this.doingAction = false;
    }
    static buildPreinitializedHook(filters) {
        const normalized = {};
        for (const tag in filters) {
            const filter = filters[tag];
            if (utils.isObject(filter) && filter instanceof POHook) {
                normalized[tag] = filter;
                continue;
            }
            const hook = new POHook();
            for (const priority in filter) {
                const cbs = filter[priority];
                cbs.forEach((cb) => {
                    hook.addFilter(cb.func, priority, cb.acceptedArgs);
                });
            }
            normalized[tag] = hook;
        }
        return normalized;
    }
    get filters() {
        return this.callbacks;
    }
    /**
     * Check if any filter has been registered
     * @since
     *
     */
    hasFilter(functionToCheck = false) {
        const funcs = Object.entries(this.callbacks).reduce((prev, [_priority, callbacks]) => {
            return prev.concat(callbacks);
        }, []);
        if (typeof functionToCheck === 'boolean') {
            return funcs.length > 0;
        }
        else {
            const cb = funcs.find(({ func }) => func === functionToCheck);
            return cb ? cb.func : false;
        }
    }
    /**
     * Add filter
     * @since
     *
     */
    addFilter(callback, priority = 10, acceptedArgs = 1) {
        if (!utils.hasOwn(this.callbacks, priority)) {
            this.callbacks[priority] = [];
        }
        this.callbacks[priority].push({
            func: callback,
            acceptedArgs,
        });
    }
    /**
     * Remove filter
     * @since
     *
     */
    removeFilter(functionToRemove, priority = 10) {
        if (utils.hasOwn(this.callbacks, priority)) {
            const index = this.callbacks[priority].findIndex(({ func }) => func === functionToRemove);
            index >= 0 && this.callbacks[priority].splice(index, 1);
            return index >= 0;
        }
        else {
            return false;
        }
    }
    /**
     * Remove filter
     * @since
     *
     */
    removeAllFilters(priority = false) {
        if (typeof priority === 'boolean') {
            this.callbacks = {};
        }
        else if (utils.hasOwn(this.callbacks, priority)) {
            this.callbacks[priority] = [];
        }
        else {
            warning_1(process.env.NDE_ENV === 'production', `priority "${priority}" is not exists , no filiters will be removed.`);
        }
    }
    applyFilters(value, ...args) {
        const sortFuncs = Object.entries(this.callbacks)
            .sort(([a], [b]) => Number(a) - Number(b))
            .reduce((prev, [_priority, callbacks]) => prev.concat(callbacks), []);
        if (sortFuncs.length) {
            const _doingAction = this.doingAction;
            return new Promise((resolve) => exec(0, value, resolve));
            // eslint-disable-next-line no-inner-declarations
            function exec(index, value, resolve) {
                const cb = sortFuncs[index];
                if (!_doingAction) {
                    args.unshift(value);
                }
                const result = cb.func.apply(null, args.slice(0, cb.acceptedArgs));
                if (index === sortFuncs.length - 1) {
                    result instanceof Promise ? result.then(resolve) : resolve(result);
                }
                else if (result instanceof Promise) {
                    result.then((val) => exec(index + 1, val, resolve));
                }
                else {
                    exec(index + 1, result, resolve);
                }
            }
        }
        return Promise.resolve(value);
    }
    doAction(...args) {
        this.doingAction = true;
        return new Promise((resolve, reject) => {
            this.applyFilters('', ...args).then(() => {
                //
                this.doingAction = false;
                resolve();
            }, reject);
        });
    }
}

const globalFilters = POHook.buildPreinitializedHook({});
function hook(tag, functionToAdd, priority, acceptedArgs) {
    if (functionToAdd) {
        if (!utils.hasOwn(globalFilters, tag)) {
            globalFilters[tag] = new POHook();
        }
        globalFilters[tag].addFilter(functionToAdd, priority, acceptedArgs);
    }
    else {
        return {
            has(functionToCheck) {
                if (!utils.hasOwn(globalFilters, tag)) {
                    return false;
                }
                return globalFilters[tag].hasFilter(functionToCheck);
            },
            remove(functionToRemove, priority = 10) {
                let removed = false;
                if (utils.hasOwn(globalFilters, tag)) {
                    removed = globalFilters[tag].removeFilter(functionToRemove, priority);
                }
                return removed;
            },
            removeAll(priority) {
                if (utils.hasOwn(globalFilters, tag)) {
                    globalFilters[tag].removeAllFilters(priority);
                }
            },
            filter(value, ...args) {
                if (utils.hasOwn(globalFilters, 'all')) ;
                if (utils.hasOwn(globalFilters, tag)) {
                    return globalFilters[tag].applyFilters(value, ...args);
                }
                return Promise.resolve(value);
            },
            exec(...args) {
                if (utils.hasOwn(globalFilters, tag)) {
                    return globalFilters[tag].doAction(...args);
                }
                return Promise.resolve();
            },
        };
    }
}

const install$1 = ModuleLoader.install;
ModuleLoader.install = (Vue) => {
    install$1.call(ModuleLoader, Vue);
};
// Auto install if it is not done yet and `window` has `Vue`.
// To allow users to avoid auto-installation in some cases,
if (typeof window !== 'undefined' && window.Vue) {
    window.Vue.use(ModuleLoader);
}

exports.default = ModuleLoader;
exports.hook = hook;
